from __future__ import annotations

import argparse
import csv
import importlib.util
import io
import json
import pickle
import urllib.request
import warnings
from contextlib import redirect_stdout
from dataclasses import dataclass
from pathlib import Path

import lls
import numpy as np
import kinepet


DEFAULT_PYPET_URL = (
    "https://raw.githubusercontent.com/Kuttner/pyPET/"
    "8d552908f5f2accb2d026f6c3081dbf5bb961096/Tissue_compartment_modeling.py"
)


@dataclass
class SampleCurves:
    sample_id: str
    npz_path: Path
    voi_path: Path
    group: str | None
    npz_sample_id: str | None
    time_minutes: np.ndarray
    aif: np.ndarray
    brain: np.ndarray
    myocardium: np.ndarray


@dataclass
class SkippedSample:
    sample_id: str
    npz_path: str
    voi_path: str
    reason: str


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=(
            "Match raw NPZ files to expert VOI pickles, extract Brain/Myocardium "
            "curves, and fit them with the PyPET reference code, the TAC-based solver, "
            "and the LLS implementation."
        )
    )
    parser.add_argument(
        "--raw-npz-glob",
        default="data/raw/*.npz",
    )
    parser.add_argument(
        "--voi-glob",
        default="/Users/csa047/uit/phd/datasets/pet/VOI_SUV/VOI_*.pkl",
    )
    parser.add_argument("--output-dir", type=Path, default=Path("output/validation/voi_reference"))
    parser.add_argument("--threads", type=int, default=1)
    parser.add_argument("--pypet-module", type=Path, default=None)
    parser.add_argument("--pypet-url", default=DEFAULT_PYPET_URL)
    return parser.parse_args()


def discover_pickles(pattern: str) -> list[Path]:
    import glob

    return sorted(Path(path) for path in glob.glob(pattern))


def sanitize_sample_id(source_path: Path) -> str:
    name = source_path.stem
    if name.startswith("VOI_"):
        return name[len("VOI_") :]
    return name


def load_paired_sample(npz_path: Path, voi_path: Path) -> SampleCurves:
    with np.load(npz_path, allow_pickle=True) as arr:
        if "time" not in arr or "aif" not in arr:
            raise KeyError("Missing NPZ keys: time or aif")
        time_minutes = np.asarray(arr["time"], dtype=np.float64)
        aif = np.asarray(arr["aif"], dtype=np.float64)
        npz_sample_id = str(arr["sample_id"].item()) if "sample_id" in arr else None
        group = str(arr["group"].item()) if "group" in arr else None

    with voi_path.open("rb") as handle:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            loaded = pickle.load(handle)

    if "VOI_t" not in loaded or "VOI_A" not in loaded:
        raise KeyError("Missing VOI_t or VOI_A")

    voi_a = loaded["VOI_A"]
    missing = [name for name in ["Brain", "Myocardium"] if name not in voi_a]
    if missing:
        raise KeyError(f"Missing VOI_A keys: {', '.join(missing)}")

    voi_t = np.asarray(loaded["VOI_t"], dtype=np.float64)
    brain = np.asarray(voi_a["Brain"], dtype=np.float64)
    myocardium = np.asarray(voi_a["Myocardium"], dtype=np.float64)

    if time_minutes.ndim != 1 or aif.ndim != 1 or voi_t.ndim != 1 or brain.ndim != 1 or myocardium.ndim != 1:
        raise ValueError("Expected 1D time, aif, and VOI curves")
    if not (time_minutes.shape == aif.shape == brain.shape == myocardium.shape == voi_t.shape):
        raise ValueError("NPZ time/aif and VOI curves do not have matching lengths")

    if np.max(time_minutes) > 60.0:
        time_minutes = time_minutes / 60.0
    if np.max(voi_t) > 60.0:
        voi_t = voi_t / 60.0
    if not np.allclose(time_minutes, voi_t, atol=5e-4, rtol=0.0):
        raise ValueError("NPZ time vector and VOI_t do not match")

    return SampleCurves(
        sample_id=sanitize_sample_id(npz_path),
        npz_path=npz_path.resolve(),
        voi_path=voi_path.resolve(),
        group=group,
        npz_sample_id=npz_sample_id,
        time_minutes=time_minutes,
        aif=aif,
        brain=brain,
        myocardium=myocardium,
    )


def ensure_pypet_module(args: argparse.Namespace) -> Path:
    if args.pypet_module is not None:
        if not args.pypet_module.exists():
            raise SystemExit(f"Missing PyPET module: {args.pypet_module}")
        return args.pypet_module

    cache_dir = args.output_dir / "reference_cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    module_path = cache_dir / "Tissue_compartment_modeling.py"
    if not module_path.exists():
        with urllib.request.urlopen(args.pypet_url) as response:
            module_path.write_bytes(response.read())
    return module_path


def load_module(module_path: Path):
    spec = importlib.util.spec_from_file_location("pypet_reference", module_path)
    if spec is None or spec.loader is None:
        raise RuntimeError(f"Could not load module from {module_path}")
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def fit_with_pypet(module, sample: SampleCurves, region: str, model: str) -> dict[str, object]:
    ct = sample.brain if region == "Brain" else sample.myocardium
    cp = sample.aif
    time = sample.time_minutes

    func = module.twoTCMrev if model == "rev" else module.twoTCMirrev
    try:
        with redirect_stdout(io.StringIO()):
            params, ki, _fit_curve, fit_error = func(
                cp.copy(),
                time.copy(),
                ct.copy(),
                INTERPOLATE=True,
                TIME_FRAMES=2.5,
                fit_vB=True,
            )
        params = np.asarray(params, dtype=np.float64)
        if model == "rev":
            K1, k2, vB, k3, k4 = params
        else:
            K1, k2, vB, k3 = params
            k4 = 0.0
        return {
            "sample_id": sample.sample_id,
            "npz_sample_id": sample.npz_sample_id or "",
            "group": sample.group or "",
            "region": region,
            "model": model,
            "method": "pypet",
            "K1": float(K1),
            "k2": float(k2),
            "k3": float(k3),
            "k4": float(k4),
            "vB": float(vB),
            "Ki": float(ki),
            "error_value": float(fit_error),
            "error_kind": "rss_interp_grid",
            "status": "success",
            "iterations": "",
            "npz_path": str(sample.npz_path),
            "voi_path": str(sample.voi_path),
        }
    except Exception as exc:
        return {
            "sample_id": sample.sample_id,
            "npz_sample_id": sample.npz_sample_id or "",
            "group": sample.group or "",
            "region": region,
            "model": model,
            "method": "pypet",
            "K1": np.nan,
            "k2": np.nan,
            "k3": np.nan,
            "k4": np.nan,
            "vB": np.nan,
            "Ki": np.nan,
            "error_value": np.nan,
            "error_kind": "rss_interp_grid",
            "status": f"error: {exc}",
            "iterations": "",
            "npz_path": str(sample.npz_path),
            "voi_path": str(sample.voi_path),
        }


def lls_error_rows(sample: SampleCurves, model: str, message: str) -> list[dict[str, object]]:
    rows: list[dict[str, object]] = []
    for region in ["Brain", "Myocardium"]:
        rows.append(
            {
                "sample_id": sample.sample_id,
                "npz_sample_id": sample.npz_sample_id or "",
                "group": sample.group or "",
                "region": region,
                "model": model,
                "method": "lls",
                "K1": np.nan,
                "k2": np.nan,
                "k3": np.nan,
                "k4": np.nan,
                "vB": np.nan,
                "Ki": np.nan,
                "error_value": np.nan,
                "error_kind": "mse_interp_grid",
                "status": f"error: {message}",
                "iterations": "",
                "npz_path": str(sample.npz_path),
                "voi_path": str(sample.voi_path),
            }
        )
    return rows


def fit_with_lls(sample: SampleCurves, model: str) -> list[dict[str, object]]:
    tacs = np.stack([sample.brain, sample.myocardium], axis=1)
    model_name = "4k_vB" if model == "rev" else "3k_vB"

    try:
        kinetic_params, _fitted_tacs, mse = lls.fit_compartment_model(
            tacs=tacs,
            aif=sample.aif,
            time=sample.time_minutes,
            model=model_name,
            time_unit="min",
            interpolate=True,
            frame_duration=2.5 / 60.0,
        )
    except Exception as exc:
        return lls_error_rows(sample, model, str(exc))

    params = np.asarray(kinetic_params, dtype=np.float64)
    mse = np.asarray(mse, dtype=np.float64).reshape(-1)
    if params.ndim != 2 or params.shape[1] != 2:
        return lls_error_rows(
            sample,
            model,
            f"Unexpected kinetic parameter shape: {tuple(params.shape)}",
        )
    if mse.shape != (2,):
        return lls_error_rows(sample, model, f"Unexpected mse shape: {tuple(mse.shape)}")

    rows: list[dict[str, object]] = []
    for region, voxel in [("Brain", 0), ("Myocardium", 1)]:
        if model == "rev":
            K1, k2, k3, k4, vB, Ki = params[:, voxel]
        else:
            K1, k2, k3, vB, Ki = params[:, voxel]
            k4 = 0.0
        rows.append(
            {
                "sample_id": sample.sample_id,
                "npz_sample_id": sample.npz_sample_id or "",
                "group": sample.group or "",
                "region": region,
                "model": model,
                "method": "lls",
                "K1": float(K1),
                "k2": float(k2),
                "k3": float(k3),
                "k4": float(k4),
                "vB": float(vB),
                "Ki": float(Ki),
                "error_value": float(mse[voxel]),
                "error_kind": "mse_interp_grid",
                "status": "success",
                "iterations": "",
                "npz_path": str(sample.npz_path),
                "voi_path": str(sample.voi_path),
            }
        )
    return rows


def cpp_error_rows(sample: SampleCurves, model: str, message: str) -> list[dict[str, object]]:
    rows: list[dict[str, object]] = []
    for region in ["Brain", "Myocardium"]:
        rows.append(
            {
                "sample_id": sample.sample_id,
                "npz_sample_id": sample.npz_sample_id or "",
                "group": sample.group or "",
                "region": region,
                "model": model,
                "method": "cpp",
                "K1": np.nan,
                "k2": np.nan,
                "k3": np.nan,
                "k4": np.nan,
                "vB": np.nan,
                "Ki": np.nan,
                "error_value": np.nan,
                "error_kind": "rmse_native_grid",
                "status": f"error: {message}",
                "iterations": "",
                "npz_path": str(sample.npz_path),
                "voi_path": str(sample.voi_path),
            }
        )
    return rows


def run_cpp_fit(args: argparse.Namespace, sample: SampleCurves, model: str) -> list[dict[str, object]]:
    tacs = np.stack([sample.brain, sample.myocardium], axis=1)
    try:
        fit = kinepet.fit_tacs(
            tacs,
            sample.time_minutes,
            sample.aif,
            model=model,
            fit_vb=True,
            threads=args.threads,
        )
    except Exception as exc:
        return cpp_error_rows(sample, model, str(exc))

    k4 = fit.k4 if fit.k4 is not None else np.zeros_like(fit.k3)
    rows: list[dict[str, object]] = []
    for region, voxel in [("Brain", 0), ("Myocardium", 1)]:
        rows.append(
            {
                "sample_id": sample.sample_id,
                "npz_sample_id": sample.npz_sample_id or "",
                "group": sample.group or "",
                "region": region,
                "model": model,
                "method": "cpp",
                "K1": float(fit.K1[voxel]),
                "k2": float(fit.k2[voxel]),
                "k3": float(fit.k3[voxel]),
                "k4": float(k4[voxel]),
                "vB": float(fit.vB[voxel]),
                "Ki": float(fit.Ki[voxel]),
                "error_value": float(fit.rmse[voxel]),
                "error_kind": "rmse_native_grid",
                "status": str(int(fit.status[voxel])),
                "iterations": str(int(fit.iterations[voxel])),
                "npz_path": str(sample.npz_path),
                "voi_path": str(sample.voi_path),
            }
        )
    return rows


def save_curves_npz(output_dir: Path, samples: list[SampleCurves]) -> Path:
    output_path = output_dir / "voi_reference_curves.npz"
    np.savez(
        output_path,
        sample_id=np.array([sample.sample_id for sample in samples], dtype=object),
        npz_path=np.array([str(sample.npz_path) for sample in samples], dtype=object),
        voi_path=np.array([str(sample.voi_path) for sample in samples], dtype=object),
        group=np.array([sample.group or "" for sample in samples], dtype=object),
        npz_sample_id=np.array([sample.npz_sample_id or "" for sample in samples], dtype=object),
        time_minutes=np.stack([sample.time_minutes for sample in samples], axis=0),
        aif=np.stack([sample.aif for sample in samples], axis=0),
        brain=np.stack([sample.brain for sample in samples], axis=0),
        myocardium=np.stack([sample.myocardium for sample in samples], axis=0),
    )
    return output_path


def save_results_csv(output_dir: Path, rows: list[dict[str, object]]) -> Path:
    output_path = output_dir / "voi_reference_fits.csv"
    fieldnames = [
        "sample_id",
        "npz_sample_id",
        "group",
        "region",
        "model",
        "method",
        "K1",
        "k2",
        "k3",
        "k4",
        "vB",
        "Ki",
        "error_value",
        "error_kind",
        "status",
        "iterations",
        "npz_path",
        "voi_path",
    ]
    with output_path.open("w", encoding="utf-8", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(rows)
    return output_path


def save_skipped_csv(output_dir: Path, skipped: list[SkippedSample]) -> Path:
    output_path = output_dir / "voi_reference_skipped.csv"
    with output_path.open("w", encoding="utf-8", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=["sample_id", "npz_path", "voi_path", "reason"])
        writer.writeheader()
        for row in skipped:
            writer.writerow(
                {
                    "sample_id": row.sample_id,
                    "npz_path": row.npz_path,
                    "voi_path": row.voi_path,
                    "reason": row.reason,
                }
            )
    return output_path


def main() -> None:
    args = parse_args()
    args.output_dir.mkdir(parents=True, exist_ok=True)

    pypet_module_path = ensure_pypet_module(args)
    pypet_module = load_module(pypet_module_path)

    npz_paths = discover_pickles(args.raw_npz_glob)
    voi_paths = discover_pickles(args.voi_glob)
    if not npz_paths:
        raise SystemExit(f"No raw NPZ files matched: {args.raw_npz_glob}")
    if not voi_paths:
        raise SystemExit(f"No VOI pickle files matched: {args.voi_glob}")

    npz_by_id = {sanitize_sample_id(path): path for path in npz_paths}
    voi_by_id = {sanitize_sample_id(path): path for path in voi_paths}

    samples: list[SampleCurves] = []
    skipped: list[SkippedSample] = []

    for sample_id, npz_path in sorted(npz_by_id.items()):
        voi_path = voi_by_id.get(sample_id)
        if voi_path is None:
            skipped.append(
                SkippedSample(
                    sample_id=sample_id,
                    npz_path=str(npz_path.resolve()),
                    voi_path="",
                    reason="missing VOI pickle",
                )
            )
            continue
        try:
            samples.append(load_paired_sample(npz_path, voi_path))
        except Exception as exc:
            skipped.append(
                SkippedSample(
                    sample_id=sample_id,
                    npz_path=str(npz_path.resolve()),
                    voi_path=str(voi_path.resolve()),
                    reason=str(exc),
                )
            )

    for sample_id, voi_path in sorted(voi_by_id.items()):
        if sample_id in npz_by_id:
            continue
        skipped.append(
            SkippedSample(
                sample_id=sample_id,
                npz_path="",
                voi_path=str(voi_path.resolve()),
                reason="missing raw NPZ",
            )
        )

    rows: list[dict[str, object]] = []

    for index, sample in enumerate(samples, start=1):
        print(f"[{index}/{len(samples)}] {sample.sample_id}")
        for model in ["irr", "rev"]:
            rows.extend(run_cpp_fit(args, sample, model))
            rows.extend(fit_with_lls(sample, model))
            for region in ["Brain", "Myocardium"]:
                rows.append(fit_with_pypet(pypet_module, sample, region, model))

    curves_path = save_curves_npz(args.output_dir, samples)
    csv_path = save_results_csv(args.output_dir, rows)
    skipped_path = save_skipped_csv(args.output_dir, skipped)
    metadata_path = args.output_dir / "metadata.json"
    metadata = {
        "raw_npz_glob": args.raw_npz_glob,
        "voi_glob": args.voi_glob,
        "raw_npz_count": len(npz_paths),
        "voi_pickle_count": len(voi_paths),
        "matched_sample_count": len(samples),
        "sample_count_fitted": len(samples),
        "sample_count_skipped": len(skipped),
        "curves_npz": str(curves_path.resolve()),
        "fits_csv": str(csv_path.resolve()),
        "skipped_csv": str(skipped_path.resolve()),
        "pypet_module": str(pypet_module_path.resolve()),
        "pypet_url": args.pypet_url if args.pypet_module is None else None,
        "cpp_solver": "kinepet Python binding",
        "cpp_threads": args.threads,
        "cpp_fit_vb": 1,
        "lls_package": str(Path(lls.__file__).resolve()),
        "lls_model_map": {"irr": "3k_vB", "rev": "4k_vB"},
        "lls_interpolate": True,
        "lls_frame_duration_seconds": 2.5,
        "pypet_interpolate": True,
        "pypet_time_frames_seconds": 2.5,
        "aif_source": "raw_npz.aif",
        "target_regions": ["Brain", "Myocardium"],
    }
    with metadata_path.open("w", encoding="utf-8") as handle:
        json.dump(metadata, handle, indent=2)

    print(f"Wrote curves: {curves_path}")
    print(f"Wrote fits: {csv_path}")
    print(f"Wrote skipped: {skipped_path}")
    print(f"Wrote metadata: {metadata_path}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
from pathlib import Path
from typing import Literal

import numpy as np
import pandas as pd

from kinepet import fit_tacs

try:
    from scripts.kinetics import Irreversible2TCMVB, Reversible2TCMVB
except ModuleNotFoundError:  # pragma: no cover - direct script execution
    from kinetics import Irreversible2TCMVB, Reversible2TCMVB


ModelName = Literal["irr", "rev"]


TIME_PROTOCOL = ((1, 30), (24, 5), (9, 20), (8, 300))
DT_S = 1.0
TINJ_S = 0.5


def feng_aif(t, tau, A1, A2, A3, L1, L2, L3):
    cp = (
        (A1 * (t - tau) - A2 - A3) * np.exp(L1 * (t - tau))
        + A2 * np.exp(L2 * (t - tau))
        + A3 * np.exp(L3 * (t - tau))
    )
    cp[: np.sum(t < tau)] = 0
    return cp


def frame_average(y: np.ndarray, t_s: np.ndarray, frame_durations_s: np.ndarray) -> np.ndarray:
    edges = np.concatenate([[0.0], np.cumsum(frame_durations_s)])
    out = np.empty(len(frame_durations_s), dtype=np.float64)
    for i, (a, b) in enumerate(zip(edges[:-1], edges[1:])):
        m = (t_s >= a) & (t_s < b)
        if np.any(m):
            out[i] = np.trapezoid(y[m], t_s[m]) / (b - a)
        else:
            out[i] = np.interp(0.5 * (a + b), t_s, y)
    return out


def generate_aif(
    t_fine_min: np.ndarray,
    t_fine_s: np.ndarray,
    frame_durations_s: np.ndarray,
    dt_s: float,
    *,
    tinj_s: float,
    tau: float,
    A1: float,
    r2: float,
    r3: float,
    L1: float,
    L2: float,
    L3: float,
) -> np.ndarray:
    k = max(1, int(round(tinj_s / dt_s)))
    kernel = np.ones(k, dtype=np.float64) / k
    A2, A3 = r2 * A1, r3 * A1

    cp = feng_aif(t_fine_min, tau, A1, A2, A3, L1, L2, L3)
    cp_inj = np.convolve(cp, kernel, mode="same")
    return frame_average(cp_inj, t_fine_s, frame_durations_s)


def generate_time(frame_protocol: tuple[tuple[int, int], ...], dt_s: float) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    frame_durations_s = np.concatenate([np.full(n_frames, frame_dur, dtype=np.float64) for n_frames, frame_dur in frame_protocol])
    edges_s = np.concatenate([[0.0], np.cumsum(frame_durations_s)])
    mids_s = 0.5 * (edges_s[:-1] + edges_s[1:])

    t_fine_s = np.arange(0.0, edges_s[-1] + dt_s, dt_s, dtype=np.float64)
    t_fine_min = t_fine_s / 60.0
    time_min = mids_s / 60.0
    return t_fine_min, t_fine_s, time_min


def generate_curves(df: pd.DataFrame, dt_s: float, tinj_s: float, model: ModelName) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    compartment_model = Reversible2TCMVB() if model == "rev" else Irreversible2TCMVB()

    t_fine_min, t_fine_s, time_min = generate_time(TIME_PROTOCOL, dt_s)
    frame_durations_s = np.concatenate(
        [np.full(n_frames, frame_dur, dtype=np.float64) for n_frames, frame_dur in TIME_PROTOCOL]
    )

    n = len(df)
    T = len(time_min)
    aif = np.empty((T, n), dtype=np.float64)
    tacs = np.empty((T, n), dtype=np.float64)

    for j, row in enumerate(df.itertuples(index=False)):
        A1m = float(row.A1m)
        A2m = float(row.A2m)
        A3m = float(row.A3m)
        curve_aif = generate_aif(
            t_fine_min,
            t_fine_s,
            frame_durations_s,
            dt_s,
            tinj_s=tinj_s,
            tau=float(row.TAUm),
            A1=A1m,
            r2=A2m / A1m,
            r3=A3m / A1m,
            L1=float(row.L1m),
            L2=float(row.L2m),
            L3=float(row.L3m),
        )
        aif[:, j] = curve_aif

        if model == "rev":
            tacs[:, j] = compartment_model(
                curve_aif,
                time_min,
                K1=np.array([float(row.K1)]),
                k2=np.array([float(row.k2)]),
                k3=np.array([float(row.k3)]),
                k4=np.array([float(row.k4)]),
                vB=np.array([float(row.vB)]),
            ).reshape(-1)
        else:
            tacs[:, j] = compartment_model(
                curve_aif,
                time_min,
                K1=np.array([float(row.K1)]),
                k2=np.array([float(row.k2)]),
                k3=np.array([float(row.k3)]),
                vB=np.array([float(row.vB)]),
            ).reshape(-1)

    return tacs, aif, time_min


def main() -> None:
    parser = argparse.ArgumentParser(description="Generate TACs from CSV and fit them")
    parser.add_argument("INPUT_FILE", type=Path, help="Path to the input CSV file")
    parser.add_argument("OUTPUT_FILE", type=Path, help="Path to the output CSV file")
    parser.add_argument("--model", choices=["irr", "rev"], default="rev", help="Compartment model to use for fitting")
    parser.add_argument("--fit-vb", action="store_true", help="Whether to fit vB or keep it fixed")
    parser.add_argument("--fixed-vb", type=float, default=0.05, help="Value to use for fixed vB (ignored if --fit-vb is set)")
    parser.add_argument("--threads", type=int, default=0, help="Number of threads to use for fitting (default: all available)")
    parser.add_argument("--max-iterations", type=int, default=40, help="Maximum number of iterations for fitting")
    parser.add_argument("--csv-sep", type=str, default=";", help="Separator used in the input CSV file (default: ';')")
    args = parser.parse_args()

    df = pd.read_csv(args.INPUT_FILE, sep=args.csv_sep)

    model: ModelName = args.model

    tacs, aifs, times = generate_curves(df, dt_s=DT_S, tinj_s=TINJ_S, model=model)

    fit = fit_tacs(
        tacs,
        times,
        aifs,
        model=model,
        fit_vb=args.fit_vb,
        fixed_vb=args.fixed_vb,
        threads=args.threads,
        max_iterations=args.max_iterations,
    )

    out_df = df.copy()
    out_df["K1_fit"] = fit.K1
    out_df["k2_fit"] = fit.k2
    out_df["k3_fit"] = fit.k3
    if model == "rev":
        out_df["k4_fit"] = fit.k4
    out_df["vB_fit"] = fit.vB
    out_df["Ki_fit"] = fit.Ki
    out_df["rmse_fit"] = fit.rmse
    out_df["iterations_fit"] = fit.iterations
    out_df["status_fit"] = fit.status

    out_df.to_csv(args.OUTPUT_FILE, index=False, sep=args.csv_sep)

if __name__ == "__main__":
    main()



from __future__ import annotations

import numpy as np
from numpy.testing import assert_allclose

import kinepet


def exp_conv_step(z: float, rate: float, dt: float, c0: float, c1: float) -> float:
    if dt <= 0.0:
        return z
    if abs(rate) < 1e-12:
        return z + 0.5 * (c0 + c1) * dt
    decay = np.exp(-rate * dt)
    i0 = (1.0 - decay) / rate
    i1 = (1.0 - decay * (1.0 + rate * dt)) / (rate**2)
    slope = (c1 - c0) / dt
    return decay * z + c1 * i0 - slope * i1


def forward_irrev(
    time: np.ndarray,
    aif: np.ndarray,
    k1: float,
    k2: float,
    k3: float,
    vb: float,
) -> np.ndarray:
    lam = max(k2 + k3, 1e-12)
    z = 0.0
    integral_cp = 0.0
    out = np.zeros_like(time)
    if time[0] > 0.0:
        z = exp_conv_step(0.0, lam, float(time[0]), 0.0, float(aif[0]))
        integral_cp = 0.5 * float(aif[0]) * float(time[0])
    out[0] = (1.0 - vb) * (k1 / lam * (k3 * integral_cp + k2 * z)) + vb * aif[0]
    for i in range(len(time) - 1):
        dt = float(time[i + 1] - time[i])
        z = exp_conv_step(z, lam, dt, float(aif[i]), float(aif[i + 1]))
        integral_cp += 0.5 * float(aif[i] + aif[i + 1]) * dt
        out[i + 1] = (1.0 - vb) * (k1 / lam * (k3 * integral_cp + k2 * z)) + vb * aif[i + 1]
    return out


def make_time() -> np.ndarray:
    return np.array(
        [0.0, 0.0833, 0.1667, 0.25, 0.3333, 0.5, 0.75, 1.0, 1.5, 2.0, 3.0, 5.0, 8.0, 12.0, 20.0, 30.0, 45.0],
        dtype=np.float64,
    )


def test_fit_tacs_recovers_irreversible_batch_with_per_tac_aif() -> None:
    time = make_time()
    time_2 = time * 1.05
    time_matrix = np.stack([time, time_2], axis=1)
    aif = np.stack(
        [
            12.0 * np.exp(-time / 0.3) + 1.5 * np.exp(-time / 6.0),
            8.5 * np.exp(-time_2 / 0.45) + 2.2 * np.exp(-time_2 / 8.0),
        ],
        axis=1,
    )
    truth = {
        "K1": np.array([0.10, 0.14], dtype=np.float64),
        "k2": np.array([0.12, 0.18], dtype=np.float64),
        "k3": np.array([0.06, 0.08], dtype=np.float64),
        "vB": np.array([0.04, 0.06], dtype=np.float64),
    }
    tacs = np.stack(
        [
            forward_irrev(time, aif[:, 0], truth["K1"][0], truth["k2"][0], truth["k3"][0], truth["vB"][0]),
            forward_irrev(time_2, aif[:, 1], truth["K1"][1], truth["k2"][1], truth["k3"][1], truth["vB"][1]),
        ],
        axis=1,
    )

    fit = kinepet.fit_tacs(tacs, time_matrix, aif, model="irr", threads=2)

    assert len(fit) == 2
    assert fit.fitted_tacs == 2
    assert_allclose(fit.K1, truth["K1"], atol=5e-3)
    assert_allclose(fit.k2, truth["k2"], atol=5e-3)
    assert_allclose(fit.k3, truth["k3"], atol=5e-3)
    assert_allclose(fit.vB, truth["vB"], atol=5e-3)
    assert fit.k4 is None
    assert fit.status_code.shape == fit.status.shape
    assert all(isinstance(value, str) for value in fit.status.tolist())
    assert np.all(np.isfinite(fit.rmse))


def test_fit_tacs_interprets_long_scan_times_as_minutes() -> None:
    time = np.array(
        [0.0, 2.0, 5.0, 10.0, 20.0, 35.0, 50.0, 70.0, 90.0, 120.0],
        dtype=np.float64,
    )
    aif = 9.0 * np.exp(-time / 4.0) + 2.0 * np.exp(-time / 45.0)
    truth = {
        "K1": 0.11,
        "k2": 0.13,
        "k3": 0.04,
        "vB": 0.05,
    }
    tac = kinepet.evaluate_model(time, aif, model="irr", **truth)
    tacs = tac[:, None]

    fit = kinepet.fit_tacs(tacs, time, aif, model="irr", max_iterations=120)

    assert_allclose(fit.K1[0], truth["K1"], atol=5e-3)
    assert_allclose(fit.k2[0], truth["k2"], atol=5e-3)
    assert_allclose(fit.k3[0], truth["k3"], atol=5e-3)
    assert_allclose(fit.vB[0], truth["vB"], atol=5e-3)
    assert_allclose(
        kinepet.default_frame_weights(time),
        np.sqrt([2.0, 2.5, 4.0, 7.5, 12.5, 15.0, 17.5, 20.0, 25.0, 30.0]),
    )


def test_fit_tacs_requires_per_tac_time_for_per_tac_aif() -> None:
    time = make_time()
    tacs = np.zeros((time.size, 2), dtype=np.float64)
    aif = np.zeros((time.size, 2), dtype=np.float64)

    try:
        kinepet.fit_tacs(tacs, time, aif, model="irr")
    except ValueError as exc:
        assert "time must have shape (T, N)" in str(exc)
    else:
        raise AssertionError("Expected fit_tacs() to require per-TAC time for per-TAC AIFs")


def test_fit_one_tac_and_evaluate_model_roundtrip_reversible() -> None:
    time = make_time()
    aif = 10.0 * np.exp(-time / 0.35) + 1.8 * np.exp(-time / 7.0)
    truth = {
        "K1": 0.12,
        "k2": 0.16,
        "k3": 0.09,
        "k4": 0.04,
        "vB": 0.05,
        "delay": 0.0,
        "dispersion": 0.0,
    }
    tac = kinepet.evaluate_model(
        time,
        aif,
        model="rev",
        **truth,
    )

    fit = kinepet.fit_one_tac(
        tac,
        time,
        aif,
        model="rev",
        max_iterations=200,
    )

    assert fit.k4 is not None
    assert isinstance(fit.status, str)
    assert isinstance(fit.status_code, int)
    assert abs(fit.K1 - truth["K1"]) < 1e-2
    assert abs(fit.k2 - truth["k2"]) < 1e-2
    assert abs(fit.k3 - truth["k3"]) < 1e-2
    assert abs(fit.k4 - truth["k4"]) < 1e-2
    assert abs(fit.vB - truth["vB"]) < 1e-2

    dispersed = kinepet.evaluate_model(
        time,
        aif,
        model="rev",
        K1=truth["K1"],
        k2=truth["k2"],
        k3=truth["k3"],
        k4=truth["k4"],
        vB=truth["vB"],
        delay=0.03,
        dispersion=0.02,
    )
    assert dispersed.shape == time.shape


def test_default_bounds_include_optional_parameters() -> None:
    bounds = kinepet.default_bounds(
        model="rev",
        fit_vb=True,
        fit_delay=True,
        fit_dispersion=True,
    )
    assert "K1" in bounds
    assert "k4" in bounds
    assert "vB" in bounds
    assert "delay" in bounds
    assert "dispersion" in bounds


def test_status_code_to_name() -> None:
    assert kinepet.status_code_to_name(1) == "gradient_converged"
    assert kinepet.status_code_to_name(4) == "max_iterations"


def test_fit_tacs_rejects_invalid_aif_columns() -> None:
    time = make_time()
    tacs = np.zeros((time.size, 3), dtype=np.float64)
    bad_aif = np.zeros((time.size, 2), dtype=np.float64)

    try:
        kinepet.fit_tacs(tacs, time, bad_aif, model="irr")
    except ValueError as exc:
        assert "aif second dimension" in str(exc)
    else:
        raise AssertionError("Expected fit_tacs() to reject invalid aif shape")
